//  Decodage du tx20

#include <avr/io.h>
#include <avr/interrupt.h>
#include "SoftwareSerial.h"
#include <RingBuf.h>            // Locoduino fifo ring buffer
#include <TimeLib.h>  

unsigned long DUREE_ECOULEE, TOP_CHRONO;
unsigned int icr1;

bool premier_demarrage=true;
bool bucketPositionA = false;             // one of the two positions of tipping-bucket     
bool first;                               // as we want readings of the (MHz) loops only at the 0th moment 

int v_vit=111;                    // Vitesse du vent en Km/h 
int v_vit_moy=111;                // vitesse moyenne du vent
int v_vit_max=111;                // rafale maxi
int v_vit_min=111;                // "rafale" moyenne mini
int v_dir=111;                    // orientation du vent
int v_dir_moy=111;                // orientation moyenne du vent
int TensionFoisCent=111;          // tension de l'accu x 100
int interval_envoi =600;          // interval d'envoi des données à Thingspeak: dix minutes = 600 secondes.
//int dailyRain = 0;

const int TensionPin = A3;        // pin that the sensor is attached to
const int TensionMini = 1000;     // Tension mini de l'accu pour envoi des données = 1000 (10.0V)

String v_orient = "?";            // Orientation du vent


float TensionAccu=0;
float coeff_A3 = 4.01929;         // Rapport du pont résistif à l'entrée A3 (mesure tension)
                                  // DATE/TIME  [+CCLK: "00/01/16,02:31:58+00"]
#define CCLK_MSG_LEN 28           // total characters in the CCLK message
#define TIME_STR_LEN 18           // the actual number of characters for date and time
char timePreamble[] = "CCLK: \"" ;
char timeStamp[TIME_STR_LEN];     // holds the time string

RingBuf<uint16_t, 30> pile_vent;  // déclaration de la pile fifo pour les données de vitesse du vent : 30 valeurs x 2s par trame = environ 1min
RingBuf<uint16_t, 30> pile_dir;   // déclaration de la pile fifo pour les données de vitesse du vent : idem

SoftwareSerial Sim900(7,8);       // Port série soft pour discuter avec la SIM900 sur les pin 7 et 8

void setup() {
 
  Serial.begin(9600);       // Port série pour déboguage
  Sim900.begin(9600);       // Port série SIM 900
  
  pinMode(2, INPUT);        // Pin 2  = entrée TX20
  pinMode(3, INPUT);        // Pin 3  = entrée pluviomètre
  pinMode(9, OUTPUT);       // Pin 9  = Sortie pour allumer la Sim900
   
  EICRA  = _BV(ISC01);      // l'interruption INT0 doit être déclenchée sur front descendant (le montage 5V-3V est inverseur)
  EIMSK = _BV(INT0);        // active INT0
  // deuxieme interrupt
  EICRA  |= _BV(ISC10);      // l'interruption INT0 doit être déclenchée sur front descendant (le montage 5V-3V est inverseur)
  EIMSK |= _BV(INT1);        // active INT0
  //
   
  icr1 = 19500;
  TCCR1A = 0;               // met le registre complet TCCR1A à 0
  TCCR1B = 0;               // met le registre complet TCCR1B à 0
  OCR1A = icr1;
         
  TOP_CHRONO = (millis()/1000);                         // Enregistrement de l'heure de démarrage, 
  
  //Serial.println("Top chrono = " + String(TOP_CHRONO) + "/ " + String(interval_envoi) + "s");              // debug
  Serial.println("Parametrage fini");                   // debug
}
 
volatile int rx=0;                            // volatile car peuvent être écrits par l'interruption
volatile int bitcount=0;
volatile char pin;
volatile unsigned char sa,sb,sd,se;
volatile unsigned int sc,sf;
//volatile byte etat_pluvio = 0;
volatile byte dailyRain = 0;
static unsigned long last_interrupt_time = 0;
volatile unsigned long interrupt_time = 0;

void loop(){                                  //  #####  BOUCLE PRINCIPALE   #####

  char a[90];
  unsigned char chk;
 
  if (rx){                                                                // #####     reception d'une trame de l'anémomètre:     #####
    chk= ( sb + (sc&0xf) + ((sc>>4)&0xf) + ((sc>>8)&0xf) );chk&=0xf;      // calcul du checksum
    if (sa==4 && sb==se && sc==sf && sd==chk){                            // si le checksum est bon
      v_dir=sb;                                                           // maj de v_dir (direction instantanée)
      v_vit = sc * 3.6 / 10;                                              // Mise en forme de la vitesse du vent de m/s en km/h

      maj_piles();
    }
  }

TensionAccu = analogRead(TensionPin);                                             // Lire la valeur sur l'entrée
TensionAccu = (TensionAccu*5/10.24)*coeff_A3;
TensionFoisCent = TensionAccu;

if (premier_demarrage==true){                                               //&& TensionFoisCent > TensionMini){
  SendTextMessage("+33677936273", 0);
  premier_demarrage=false;
  }
 
DUREE_ECOULEE = (millis()/1000) - TOP_CHRONO;                             // GESTION DE L'envoi fréquenciel des données 

  //dailyRain =  dailyRain + etat_pluvio;
  //etat_pluvio =0;
  Serial.println(dailyRain);         
        
if (DUREE_ECOULEE > interval_envoi){
  TOP_CHRONO = millis()/1000;
  if (TensionFoisCent>TensionMini && premier_demarrage==false){                                       // Si la tension est bonne, sinon envoyer un sms
    Serial.println("Top");
    if (rx>0){
      Serial.println("Thingspeak");
      majThinkspeak(); 
      rx=0;
      dailyRain=0;     
    }
    else{
      Serial.println("Pas de donnees");
    }
    }else{
      Serial.println("SMS");
      SendTextMessage("+33677936273", 1);
      rx=0;
      dailyRain=0;
    }
  }
    Lire_Sim900();
    delay(500);
if(dailyRain>250){dailyRain=0;}
}
// ################################################   FIN DE BOUCLE PRINCIPALE   ##############################################################

void SendTextMessage(String num, int parametre)         // *********     ENVOYER UN SMS, parametre 0 pour le démarrage ou 1 pour le vent************
  {
  powerSim900(1);                                       // Allume la sim900
  //Sim900.println("AT+CNMI=1,2,0,0,0");                  // Paramètrage de la sim900: recevoir les sms directement sur le port serie,pas de stockage dans la carte sim
  //delay(1000);
  Sim900.println("AT+CMGF=1");                          // Configure le mode SMS en mode texte
  delay(1000);
  Lire_Sim900();
    
    String chaine_caract;
    
    Serial.println("Envoyer un SMS:");
    chaine_caract = "AT + CMGS = \"" + num + "\"";      
    Serial.println(chaine_caract);
    Sim900.println(chaine_caract);                      // Envoyer un sms au numéro contenant le code de pays
    delay(500);
    
    switch(parametre){                                 
      case 1:                                           // Si envoi du vent
        switch (v_dir){                                 // Mise en forme de l'orientation du vent en texte
        case 0:
        v_orient="N";
        break;
        case 1:
        v_orient="NNE";
        break;
        case 2:
        v_orient="NE";
        break;
        case 3:
        v_orient="ENE";
        break;
        case 4:
        v_orient="E";
        break;
        case 5:
        v_orient="ESE";
        break;
        case 6:
        v_orient="SE";
        break;
        case 7:
        v_orient="SSE";
        break;
        case 8:
        v_orient="S";
        break;
        case 9:
        v_orient="SSO";
        break;
        case 10:
        v_orient="SO";
        break;
        case 11:
        v_orient="OSO";
        break;
        case 12:
        v_orient="O";
        break;
        case 13:
        v_orient="ONO";
        break;
        case 14:
        v_orient="NO";
        break;
        case 15:
        v_orient="NNO";
        break;      
      }
    chaine_caract = "Moy=" + String(v_vit_moy) + " - Min=" + String(v_vit_min) + " - Max=" + String(v_vit_max) + " - Dir=" + v_orient + " -- Tension= " + String(TensionFoisCent)+ " -- Pluie= " + String(dailyRain);
    break;
    case 0:                                              // démarrage
      chaine_caract = "Demarrage, Tension= " + String(TensionFoisCent);
    break;
    }
    Serial.println(chaine_caract);                       //Contenu du message debug
    Sim900.println(chaine_caract);                       //Contenu du message
    delay(500);
    Lire_Sim900();
    Sim900.print((char)26);                             //Le code ASCII de "ctrl+z" est 26 (requis d'après la datasheet)
    //delay(500);
    //Sim900.println();
    delay(4000);
    Lire_Sim900();
    Serial.println("Message envoye.");
    powerSim900(0);                                     // Eteint la sim900
    Lire_Sim900();
    Sim900.flush();
  }

void powerSim900(int pwr)                               // ######     GESTION DE L'ALIM DE LA SIM900   ######
{
  Serial.print("power: ");
  Serial.println(pwr);  
  int state = 0;                                        // CHECK STATE OF GSM
  Sim900.flush();                                       // vide le tampon série
  delay(1000);
  Sim900.println("AT");
  delay(500);
  if (Sim900.available() > 0) { state = 1; }  
  Lire_Sim900();
  Sim900.flush(); // get rid of anything left
 
  if (((pwr == 1) && (state == 1)) || ((pwr == 0) && (state == 0))) // ne rien faire si déjà dans le bon mode
  {
    Serial.println("power: ne rien faire");
    return;
  }
  Serial.println("power: basculer");
  pinMode(9, OUTPUT);                                     // On switch le bouton  
  digitalWrite(9,LOW);
  delay(1000);
  digitalWrite(9,HIGH);
  delay(2000);
  digitalWrite(9,LOW);
  delay(8000);                                         // Donne le temps pour accrocher le réseau
  Lire_Sim900();
  Sim900.flush();                                       // get rid of anything left
}

void ecrit_vent()                                       // ##### ecrit les infos du vent sur le port série de débogage  #####
{
  Serial.print("Inst=");
  Serial.print(v_vit);
  Serial.print(" - DirInstMoy=");
  Serial.print(v_dir);  
  Serial.print(" - Moy=");
  Serial.print(v_vit_moy);
  Serial.print("- Max=");
  Serial.print(v_vit_max);
  Serial.print("- Min=");
  Serial.print(v_vit_min);
  Serial.print("- Dir=");
  Serial.print(v_dir_moy);
  Serial.println();
}

void maj_piles()
{
      uint16_t poubelle;
      if(pile_vent.isFull()) pile_vent.lockedPop(poubelle);                                // Si a pile est pleine on libere d'abords une place
      pile_vent.push(v_vit);                                                               // ajout de la valeur actuelle
      if (pile_dir.isFull()) pile_dir.lockedPop(poubelle);                                 // idem pour la direction
      pile_dir.push(v_dir);

      v_vit_moy=v_vit;                                                                     // calculs des moyennes
      v_vit_min=v_vit;
      v_vit_max=v_vit;
      v_dir_moy=v_dir;
      
      for(int i=0; i<pile_vent.size(); i++){
        v_vit_moy=v_vit_moy + pile_vent[i];
        if(v_vit_max<pile_vent[i])v_vit_max=pile_vent[i];
        if(v_vit_min>pile_vent[i])v_vit_min=pile_vent[i];        
      }
      v_vit_moy=v_vit_moy / pile_vent.size();
  
      for(int j=0; j<pile_dir.size(); j++){
        v_dir_moy=  v_dir_moy+pile_dir[j];
      }
      v_dir_moy = v_dir_moy / pile_dir.size();
}

void Lire_Sim900()                                        // Envoie ce qui vient du port serie sim900 vers le port série debug
{
  delay(500);
  while(Sim900.available()) 
  {
    Serial.write(Sim900.read());
  }
}

void majThinkspeak(){
  powerSim900(1);                       // allumer la sim900
  Sim900.println("AT");
  delay(1000);
  Sim900.println("AT+CPIN?");
  delay(1000);
  Sim900.println("AT+CREG?");
  delay(1000);
  Sim900.println("AT+CGATT?");
  delay(1000);
  Sim900.println("AT+CIPSHUT");           // fermer les communication existantes
  delay(1000);
  Sim900.println("AT+CIPSTATUS");
  Lire_Sim900();
  delay(2000);
  Sim900.println("AT+CIPMUX=0");
  delay(2000);
  Lire_Sim900();
  Sim900.println("AT+CSTT=\"websfr\",\"\",\"\"");   // sl2sfr / websfr / wapsfr : parametrage de l'APN SFR,
  delay(1000);
  Lire_Sim900();
  Sim900.println("AT+CIICR");           //Allume la connection gprs
  delay(3000);
  Lire_Sim900();
  Sim900.println("AT+CIFSR");           //récupère une adresse IP
  delay(2000);
  Lire_Sim900();
  Sim900.println("AT+CIPSPRT=0");       // était 0 : met le caractère > prompt quand le module envoie des données
  delay(3000);
  Lire_Sim900();
  Sim900.println("AT+CIPSTART=\"TCP\",\"api.thingspeak.com\",\"80\"");        //démarre la connexion TCP
  delay(6000);
  Lire_Sim900();
  Sim900.println("AT+CIPSEND");                                               //déclenche l'envoi des données au serveur distant
  delay(4000);
  Lire_Sim900();
  
  String str="GET /update?api_key=VFB792BP4L39N3NO&field1=0"+String(v_vit_min)+"&field2=0"+String(v_vit_moy)+"&field3=0"+String(v_vit_max)+"&field4=0"+String(v_dir)+"&field5=0"+String(TensionFoisCent)+"&field6=0"+String(dailyRain)+" HTTP/1.0\r\n\r\n";

  Sim900.println(str);                  //commence à envoyer les données
  delay(100);
  Sim900.println((char)26);             //caractère retour pour envoi des données
  delay(5000);  
  Lire_Sim900();                       //attendre la réponse, important! dépends de la qualité de la liaison
  Sim900.println("AT+CIPSEND?");
  delay(1000); 
  Lire_Sim900();
  Sim900.println();
  Lire_Sim900();
  Sim900.println("AT+CIPSHUT");         //ferme la connexion
  delay(2000);
  Lire_Sim900();
  powerSim900(0);                       // Eteindre la sim900
}

boolean getTimeStamp(){

 Serial.println("getTimeStamp()"); 
 Sim900.flush();
 delay(100);
 Sim900.println("AT+CCLK?");                                        //Commande AT pour recevoir le time stamp
 delay(1000);
 if (readTime() == true){ 
  int yr = (((timeStamp[0])-48)*10)+((timeStamp[1])-48);
  int month = (((timeStamp[3])-48)*10)+((timeStamp[4])-48);
  int day  = (((timeStamp[6])-48)*10)+((timeStamp[7])-48);
  int hr = (((timeStamp[9])-48)*10)+((timeStamp[10])-48);
  int min = (((timeStamp[12])-48)*10)+((timeStamp[13])-48);
  int sec = (((timeStamp[15])-48)*10)+((timeStamp[16])-48);
 
    Serial.print("DATE/TIME: ");
    Serial.print("Year:");
    Serial.println(yr, DEC);
    Serial.print("Month:");
    Serial.println(month, DEC);
    Serial.print("Hour:");
    Serial.println(hr, DEC);
    Serial.print("Min:");
    Serial.println(min, DEC);
    Serial.print("Sec:");
    Serial.println(sec, DEC);

  setTime(hr,min,sec,day,month,yr);                                 // Paramètre l'horloge soft de l'arduino
 }
 else {
   Serial.println("Cannot get current time!");
   Sim900.flush();
   delay(1000);
   return false;
 }
 Sim900.flush(); 
 delay(1000);
 return true;
}

boolean readTime(){                                                 // Attention, code non sûre...
  Serial.println("readTime(): "); 

 if(Sim900.available() < CCLK_MSG_LEN )                             // Check for available bytes
    return false;                                                   // Not enough characters for a full message.  Try again later
 char ch = Sim900.read();                                           // skip first two chars
 ch = Sim900.read();
 if(Sim900.read() != '+')
    return false;                                                   // Not in sync yet

 for(int i=0; i < strlen(timePreamble); i++) {                      // In sync.  Check the preamble
     if(Sim900.read() != timePreamble[i] )
         return false;                                              // exit if the received data does not match the preamble 
 }
 
 for(int i=0; i < TIME_STR_LEN; i++)                                // Preamble checked.  We are now at the start of the time data
       timeStamp[i] = Sim900.read();
  Serial.println(timeStamp);
 Sim900.flush();
 return true;                                                       // we have filled the timeString with valid data 
}

// ***************  INTERRUPTION POUR LIRE LA L'ANEMOMETRE ************

ISR (INT0_vect){                                        // Interruption changement état pin 2

  sa=sb=sd=se=0;                                        // initialise les accumulateurs
  sc=0;sf=0;
 
  EIMSK &= ~_BV(INT0);                                  // desactive l'interruption
 
  bitcount=41;
 
  TCCR1B = 0;                           // arrete le timer
  TCNT1 = icr1>>1;                      // prépare le compteur
  TIFR1 |= _BV(OCF1A);                  // efface l'interruption timer
  TCCR1B = _BV(WGM12)| _BV(CS10);       // demarre le timer
  TIMSK1 |= _BV(OCIE1A);                // active l'interruption timer
 
}

ISR (INT1_vect){

  interrupt_time = millis();
  // If interrupts come faster than 200ms, assume it's a bounce and ignore
  if (interrupt_time - last_interrupt_time > 200)
 {
   dailyRain =  dailyRain + 1;
 }
 last_interrupt_time = interrupt_time;
}


ISR(TIMER1_COMPA_vect){                 // timer irq ( bit accumulator )
 
  if (bitcount==1){
    TIMSK1 &= ~_BV(OCIE1A);           // disable timer interrupt
    EIFR |= _BV(INTF0);               // clear pin interrupt flag
    EIMSK = _BV(INT0);                // enable pin interrupt flag for next frame
 
  }
  if (bitcount>0){
    pin = digitalRead(2);
 
    if (bitcount > 41-5){
      // start
      sa = (sa<<1)|pin;
    } else
    if (bitcount > 41-5-4){
      // wind dir
      sb = sb>>1 | (pin<<3);
    } else
    if (bitcount > 41-5-4-12){
      // windspeed
      sc = sc>>1 | (pin<<11);
    } else
    if (bitcount > 41-5-4-12-4){
      // checksum
      sd = sd>>1 | (pin<<3);
    } else 
    if (bitcount > 41-5-4-12-4-4){
      // wind dir, inverted
      se = se>>1 | ((pin^1)<<3);
    } else {
      // windspeed, inverted
      sf = sf>>1 | ((pin^1)<<11);
    }     
 
    bitcount--;
    if (bitcount==0){                 // Données reçues
      rx=1;
    }
  } 
}
